import datetime

from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase



class TestItemFields(TestBase):
    """ Contains tests for setting field values and processing them for both
    normal items and bulk items. This class does not contain tests for
    persisting items to a database.
    """
    
    ModelFieldTypes = None
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    
    @classmethod
    def setup_models(cls, aliased=False):
        cls.item_cls_manager.clear()
        
        class ItemFieldTypes(Item):
            model_cls = cls.ModelFieldTypes
            conversions = {
                'decimal_separator': ',',
            }
        cls.ItemFieldTypes = ItemFieldTypes
        
        dct_1 = {'model_cls': cls.ModelGeneralOne}
        dct_2 = {'model_cls': cls.ModelGeneralTwo}
        
        if aliased:
            dct_1.update({
                'aliases': {
                    'alias_1': 'two_1_1__f_integer',
                    'alias_2': 'two_x_x__one_1_1__f_integer',
                }
            })
            dct_2.update({
                'aliases': {
                    'alias_1__post': 'f_string',
                    'alias_2__post': 'one_x_1__alias_1',
                    'alias_3__post': 'one_x_x__two_1_x__alias_1__post',
                }
            })
        
        cls.ItemGeneralOne = type('ItemGeneralOne', (Item,), dct_1)
        cls.ItemGeneralTwo = type('ItemGeneralTwo', (Item,), dct_2)
        
        cls.ItemGeneralOne.complete_setup()
        cls.ItemGeneralTwo.complete_setup()
        
        
    def test_setting_fields(self):
        self.setup_models()
        
        #--- setting with correct keys ---
        item = self.ItemGeneralOne()
        item['f_integer'] = '10'
        item['two_1_1__f_string'] = 'one'
        item['two_1_1__one_x_1__f_string'] = 'two'
        item['two_1_1__one_x_x__two_x_1__f_string'] = 'three'
        bulk = item['two_1_1__one_x_x']
        bulk['f_integer'] = '20'
        bulk.gen(f_text='four')
        in_bulk_item = bulk.gen(f_text='five')
        # setting field by calling
        in_bulk_item(f_string='ITEM_CALL') 
        bulk(two_x_1__f_text='BULK_CALL')
        
        expected_value = {
            'item': {
                'f_integer': '10',
                'two_1_1': {
                    'item': {
                        'f_string': 'one',
                        'one_x_1': {
                            'item': {
                                'f_string': 'two'
                            }
                        },
                        'one_x_x': {
                            'bulk': [
                                {
                                    'item': {
                                        'f_text': 'four'
                                    }
                                },
                                {
                                    'item': {
                                        'f_string': 'ITEM_CALL',
                                        'f_text': 'five'
                                    }
                                }
                            ],
                            'defaults': {
                                'f_integer': '20',
                                'two_x_1__f_string': 'three',
                                'two_x_1__f_text': 'BULK_CALL'
                            }
                        }
                    }
                }
            }
        }
        self.assertEqual(item.to_dict(), expected_value)
        
        #--- setting with incorrect keys ---
        item = self.ItemGeneralOne()
        with self.assertRaises(KeyError):
            item['wrong_key'] = 'value'
        
        bulk = self.ItemGeneralOne.Bulk()
        with self.assertRaises(KeyError):
            bulk['wrong_key'] = 'value'
        
        #--- aliases -----------------------------------------------------------
        self.setup_models(aliased=True)
        
        item = self.ItemGeneralOne()
        item['alias_1'] = '1'
        item['alias_2'] = '2'
        item['two_1_1__alias_1__post'] = 'str-1'
        item['two_1_1__alias_2__post'] = '3'
        item['two_1_1__alias_3__post'] = '4'
        bulk = item['two_1_x']
        bulk['alias_1__post'] = 'str-2'
        
        expected_value = {
            'item': {
                'two_1_1': {
                    'item': {
                        'f_integer': '1',  # alias_1
                        'f_string': 'str-1',  # two_1_1__alias_1__post
                        'one_x_1': {
                            'item': {
                                'two_1_1': {
                                    'item': {
                                        # two_1_1__alias_2__post
                                        'f_integer': '3'
                                    }
                                }
                            }
                        },
                        'one_x_x': {
                            'bulk': [],
                            'defaults': {
                                'two_1_x__f_string': '4'
                            }
                        }
                    }
                },
                'two_1_x': {
                    'bulk': [],
                    'defaults': {
                        'f_string': 'str-2'
                    }
                },
                'two_x_x': {
                    'bulk': [],
                    'defaults': {
                        'one_1_1__f_integer': '2'  # alias_1
                    }
                }
            }
        }
        self.assertEqual(item.to_dict(), expected_value)
        
    
    def test_getting_fields(self):
        self.setup_models()
        
        item = self.ItemGeneralOne()
        related_item = item['two_x_1__one_1_1__two_1_1'](f_integer='20',
                                                         f_string='value')
        self.assertIsInstance(related_item, self.ItemGeneralTwo)
        
        expected_value = {
            'item': {
                'two_x_1': {
                    'item': {
                        'one_1_1': {
                            'item': {
                                'two_1_1': {
                                    'item': {
                                        'f_integer': '20',
                                        'f_string': 'value',
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        self.assertEqual(item.to_dict(), expected_value)
        
        #--- aliases -----------------------------------------------------------
        self.setup_models(aliased=True)
        
        item = self.ItemGeneralOne()
        item['alias_1'] = '1'
        item['alias_2'] = '2'
        item['two_1_1__alias_1__post'] = 'str-1'
        item['two_1_1__alias_2__post'] = '3'
        item['two_1_1__alias_3__post'] = '4'
        bulk = item['two_1_x']
        bulk['alias_1__post'] = 'str-2'
        
        self.assertEqual(item['alias_1'], '1')
        self.assertEqual(item['alias_2'], '2')
        self.assertEqual(item['two_1_1__alias_1__post'], 'str-1')
        self.assertEqual(item['two_1_1__alias_2__post'], '3')
        self.assertEqual(item['two_1_1__alias_3__post'], '4')
        self.assertEqual(bulk['alias_1__post'], 'str-2')
        
    
    def test_del_from_item(self):
        self.setup_models()
        
        item = self.ItemGeneralOne()
        item['f_integer'] = '10'
        item['f_string'] = 'str-10'
        item['two_1_1__f_integer'] = '20'
        item['two_1_1__f_string'] = 'str-20'
        item['two_x_x__one_1_1__f_integer'] = '30'
        item['two_x_x__one_1_1__f_string'] = 'str-30'
        bulk = item['two_1_x']
        bulk['f_string'] = 'str-2'
        
        expect = {
            'item': {
                'f_integer': '10',
                'f_string': 'str-10',
                'two_1_1': {
                    'item': {
                        'f_integer': '20',
                        'f_string': 'str-20'
                    }
                },
                'two_1_x': {
                    'bulk': [],
                    'defaults': {
                        'f_string': 'str-2'
                    }
                },
                'two_x_x': {
                    'bulk': [],
                    'defaults': {
                        'one_1_1__f_integer': '30',
                        'one_1_1__f_string': 'str-30'
                    }
                }
            }
        }
        self.assertEqual(item.to_dict(), expect)
        
        del expect['item']['f_string']
        del item['f_string']
        self.assertEqual(item.to_dict(), expect)
        
        del expect['item']['two_1_1']['item']['f_string']
        del item['two_1_1__f_string']
        self.assertEqual(item.to_dict(), expect)
        
        del expect['item']['two_x_x']['defaults']['one_1_1__f_string']
        del item['two_x_x__one_1_1__f_string']
        self.assertEqual(item.to_dict(), expect)
        
        del bulk['f_string']
        del expect['item']['two_1_x']['defaults']['f_string']
        self.assertEqual(item.to_dict(), expect)
        
        #--- aliases -----------------------------------------------------------
        self.setup_models(aliased=True)
        
        item = self.ItemGeneralOne()
        item['alias_1'] = '1'
        item['alias_2'] = '2'
        item['two_1_1__alias_1__post'] = 'str-1'
        item['two_1_1__alias_2__post'] = '3'
        item['two_1_1__alias_3__post'] = '4'
        bulk = item['two_1_x']
        bulk['alias_1__post'] = 'str-2'
        
        expected_value = {
            'item': {
                'two_1_1': {
                    'item': {
                        'f_integer': '1',
                        'f_string': 'str-1',
                        'one_x_1': {
                            'item': {
                                'two_1_1': {
                                    'item': {
                                        'f_integer': '3'
                                    }
                                }
                            }
                        },
                        'one_x_x': {
                            'bulk': [],
                            'defaults': {
                                'two_1_x__f_string': '4'
                            }
                        }
                    }
                },
                'two_1_x': {
                    'bulk': [],
                    'defaults': {
                        'f_string': 'str-2'
                    }
                },
                'two_x_x': {
                    'bulk': [],
                    'defaults': {
                        'one_1_1__f_integer': '2'
                    }
                }
            }
        }
        self.assertEqual(item.to_dict(), expected_value)
        
        del item['alias_1']
        del expected_value['item']['two_1_1']['item']['f_integer']
        self.assertEqual(item.to_dict(), expected_value)
        
        del item['alias_2']
        del expected_value['item']['two_x_x']['defaults']['one_1_1__f_integer']
        self.assertEqual(item.to_dict(), expected_value)
        
        del item['two_1_1__alias_1__post']
        del expected_value['item']['two_1_1']['item']['f_string']
        self.assertEqual(item.to_dict(), expected_value)
        
        del item['two_1_1__alias_2__post']
        del expected_value['item']['two_1_1']['item']['one_x_1']['item']['two_1_1']['item']['f_integer']
        self.assertEqual(item.to_dict(), expected_value)
        
        del item['two_1_1__alias_3__post']
        del expected_value['item']['two_1_1']['item']['one_x_x']['defaults']['two_1_x__f_string']
        self.assertEqual(item.to_dict(), expected_value)
        
        del bulk['alias_1__post']
        del expected_value['item']['two_1_x']['defaults']['f_string']
        self.assertEqual(item.to_dict(), expected_value)
        
        
    def test_field_convesions(self):
        self.setup_models()
        
        #--- simple conversion ---
        item = self.ItemFieldTypes(
            binary_1 = 'binary data',
            string_1 = 1000,
            text_1 = 2000,
            integer_1 = '10',
            boolean_1 = 'TRUE',
            float_1 = '1.120,30',  # with comma as decimal separator and a dot
            date_1 = '2000-10-30',
            time_1 = '20:30:40',
            datetime_1 = '2000-10-30 20:30:40')
        
        item.process()
        expected_value = {
            'item': {
                'binary_1': b'binary data',
                'boolean_1': True,
                'date_1': datetime.date(2000, 10, 30),
                'datetime_1': datetime.datetime(2000, 10, 30, 20, 30, 40),
                'float_1': 1120.3,
                'integer_1': 10,
                'string_1': '1000',
                'text_1': '2000',
                'time_1': datetime.time(20, 30, 40)
            }
        }
        self.assertEqual(item.to_dict(), expected_value)
        
        item.process()  # second processing does nothing
        self.assertEqual(item.to_dict(), expected_value)
        
        #--- conversions with relations ---
        item = self.ItemGeneralOne(f_integer='10',
                                   two_x_1__f_integer='20',
                                   two_x_x__f_integer='30')
        item['two_x_x'].gen(f_integer='40')
        item.process()
        
        expected_value = {
            'item': {
                'f_integer': 10,
                'two_x_1': {
                    'item': {
                        'f_integer': 20
                    }
                },
                'two_x_x': {
                    'bulk': [
                        {
                            'item': {
                                'f_integer': 40
                            }
                        }
                    ],
                    'defaults': {
                        'f_integer': 30
                    }
                }
            }
        }
        self.assertEqual(item.to_dict(), expected_value)
        
    
    def test_inject_nullables(self):
        self.setup_models()
        
        self.item_cls_manager.clear()
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            # using normal fields
            nullables = ['f_integer', 'f_string']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            # using relations
            nullables = ['one_1_1', 'one_x_1', 'one_1_x', 'one_x_x']
        
        # no overwrite for normal fields
        item = ItemGeneralOne(f_integer='10', f_string='20', f_text='30')
        item.process()
        expect = {
            'item': {
                'f_integer': 10,
                'f_string': '20', 
                'f_text': '30',
            }
        }
        self.assertEqual(item.to_dict(), expect)
        
        # normal fields nullables added
        item = ItemGeneralOne()
        item.process()
        expect = {
            'item': {
                'f_integer': None,
                'f_string': None,
            }
        }
        self.assertEqual(item.to_dict(), expect)
        
        # no overwrite for relations
        item = ItemGeneralTwo(one_1_1__f_integer='10',
                              one_1_1__f_string='20',
                              one_x_1__f_integer='10',
                              one_x_1__f_string='20')
        item['one_1_x'].gen(f_integer='10', f_string='20')
        item['one_x_x'].gen(f_integer='10', f_string='20')
        item.process()
        expect = {
            'item': {
                'one_1_1': {
                    'item': {
                        'f_integer': 10,
                        'f_string': '20'
                    }
                },
                'one_1_x': {
                    'bulk': [
                        {
                            'item': {
                                'f_integer': 10,
                                'f_string': '20'
                            }
                        }
                    ],
                    'defaults': {}
                },
                'one_x_1': {
                    'item': {
                        'f_integer': 10,
                        'f_string': '20'
                    }
                },
                'one_x_x': {
                    'bulk': [
                        {
                            'item': {
                                'f_integer': 10,
                                'f_string': '20'
                            }
                        }
                    ],
                    'defaults': {}
                }
            }
        }
        self.assertEqual(item.to_dict(), expect)
        
        # relation fields nullables added
        item = ItemGeneralTwo()
        item.process()
        expect = {
            'item': {
                'one_1_1': None,
                'one_1_x': {
                    'bulk': [],
                    'defaults': {}
                },
                'one_x_1': None,
                'one_x_x': {
                    'bulk': [],
                    'defaults': {}
                }
            }
        }
        self.assertEqual(item.to_dict(), expect)
        
    
    def test_inject_bulk_defaults(self):
        self.setup_models()
        
        self.item_cls_manager.clear()
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            # using normal fields
            nullables = ['f_integer', 'f_string']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            # using relations
            nullables = ['one_1_1', 'one_1_x']
        
        # default regular fields
        # (first default fields must be injected only then nullables)
        bulk = ItemGeneralOne.Bulk(f_integer='1000', f_text='text-value')
        bulk.gen(f_integer='10', f_float='20.30')
        bulk.gen(f_string='str-value', f_float='20.30')
        bulk.process()
        expect = {
            'bulk': [
                {
                    'item': {
                        'f_float': 20.3,
                        'f_integer': 10,
                        'f_string': None,
                        'f_text': 'text-value'
                    }
                },
                {
                    'item': {
                        'f_float': 20.3,
                        'f_integer': 1000,
                        'f_string': 'str-value',
                        'f_text': 'text-value'
                    }
                }
            ],
            'defaults': {
                'f_integer': 1000,
                'f_text': 'text-value'
            }
        }
        self.assertEqual(bulk.to_dict(), expect)
        
        # default relations
        default_two_1_1 = ItemGeneralTwo(f_integer='10', f_float='20.30')
        default_two_1_x = ItemGeneralTwo.Bulk()
        default_two_1_x.gen(f_integer='20')
        default_two_1_x.gen(f_float='40.50')
        
        bulk = ItemGeneralOne.Bulk(two_1_1=default_two_1_1,
                                   two_1_x=default_two_1_x)
        bulk.gen(two_1_1__f_integer=40)
        bulk.gen(two_1_x__f_text='text-value')
        
        bulk.process()
        expect = {
            'defaults': {
                # `ItemGeneralTwo(f_integer='10', f_float='20.30')` + nullables
                'two_1_1': {
                    'f_float': 20.3,
                    'f_integer': 10,
                    'one_1_1': None,
                    'one_1_x': {
                        'bulk': [],
                        'defaults': {}
                    }
                },
                # `default_two_1_x = ItemGeneralTwo.Bulk()` with two items:
                # `default_two_1_x.gen(f_integer='20')` and
                # `default_two_1_x.gen(f_float='40.50')` + nullables
                'two_1_x': {
                    'bulk': [
                        {
                            'f_integer': 20,
                            'one_1_1': None,
                            'one_1_x': {
                                'bulk': [],
                                'defaults': {}
                            }
                        },
                        {
                            'f_float': 40.5,
                            'one_1_1': None,
                            'one_1_x': {
                                'bulk': [],
                                'defaults': {}
                            }
                        }
                    ],
                    'defaults': {}
                }
            },
            'bulk': [
                # first item: `bulk.gen(two_1_1__f_integer=40)`
                # + nullables + defaults
                {
                    'f_integer': None,
                    'f_string': None,
                    # exited relation
                    'two_1_1': {
                        'f_integer': 40,
                        'one_1_1': None,
                        'one_1_x': {
                            'bulk': [],
                            'defaults': {}
                        }
                    },
                    # added from default
                    'two_1_x': {
                        'bulk': [
                            {
                                'f_integer': 20,
                                'one_1_1': None,
                                'one_1_x': {
                                    'bulk': [],
                                    'defaults': {}
                                }
                            },
                            {
                                'f_float': 40.5,
                                'one_1_1': None,
                                'one_1_x': {
                                    'bulk': [],
                                    'defaults': {}
                                }
                            }
                        ],
                        'defaults': {}
                    }
                },
                # second item: `bulk.gen(two_1_x__f_text='text-value')`
                # + nullables + defaults
                {
                    'f_integer': None,
                    'f_string': None,
                    # added from default
                    'two_1_1': {
                        'f_float': 20.3,
                        'f_integer': 10,
                        'one_1_1': None,
                        'one_1_x': {
                            'bulk': [],
                            'defaults': {}
                        }
                    },
                    # already existed
                    'two_1_x': {
                        'bulk': [],
                        'defaults': {
                            'f_text': 'text-value'
                        }
                    }
                }
            ]
        }
        self.assertTrue(bulk.to_dict(), expect)

    
    def test_can_create(self):
        self.setup_models()
        
        self.item_cls_manager.clear()
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'two_1_1', 'f_integer', 'f_string'}]
            getters = [{'f_integer', 'f_string'}]
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
        
        # item one has creator field
        item_one = ItemGeneralOne(f_integer=10)
        self.assertFalse(item_one.can_create(fkeys={}))
        item_one['f_string'] = 'str-one'
        self.assertFalse(item_one.can_create(fkeys={}))
        
        # relation (using `True` as fake ORM model value)
        self.assertTrue(item_one.can_create(fkeys={'two_1_1': True}))
        
        # item two (empty creators)
        item_two = ItemGeneralTwo()
        # can create with all fields equal to `None`
        self.assertTrue(item_two.can_create(fkeys={}))
        item_two['f_integer'] = '1000'
        self.assertTrue(item_two.can_create(fkeys={}))
        
        # item one no fkey but has field
        item_one = ItemGeneralOne(f_integer=10, f_string='20',
                                  two_1_1=item_two)
        self.assertFalse(item_one.can_create(fkeys={}))


        # `model_cls is None`
        class TestItem(Item):
            pass
        self.assertIsNone(TestItem.model_cls)
        test_item = TestItem()
        self.assertFalse(test_item.can_create(fkeys={}))
    
    
    def test_contains(self):
        self.setup_models()
        item = self.ItemGeneralOne()
        item['f_integer'] = '1'
        item['two_1_1__f_integer'] = '2'
        item['two_1_x__f_integer'] = '3'
        bulk = item['two_1_x']
        bulk['one_1_1__f_integer'] = '4'
        bulk['one_x_x__f_integer'] = '5'
        
        self.assertIn('f_integer', item)
        self.assertIn('two_1_1__f_integer', item)
        self.assertIn('two_1_x__f_integer', item)
        self.assertIn('two_1_x__one_1_1__f_integer', item)  # from bulk
        self.assertIn('two_1_x__one_x_x__f_integer', item)  # from bulk
        
        self.assertNotIn('wrong_key', item)
        self.assertNotIn('wrong_key__f_integer', item)
        self.assertNotIn('two_1_x__one_x_x__wrong_key', item)
        
        self.assertIn('f_integer', bulk)  # item
        self.assertIn('one_1_1__f_integer', bulk)
        self.assertIn('one_x_x__f_integer', bulk)
        
        self.assertNotIn('wrong_key', bulk)
        self.assertNotIn('wrong_key__f_integer', bulk)
        self.assertNotIn('two_1_x__one_x_x__wrong_key', bulk)
        
        #--- aliased -----------------------------------------------------------
        self.setup_models(aliased=True)
        
        item = self.ItemGeneralOne()
        item['alias_1'] = '1'
        item['alias_2'] = '2'
        item['two_1_1__alias_1__post'] = 'str-1'
        item['two_1_1__alias_2__post'] = '3'
        item['two_1_1__alias_3__post'] = '4'
        bulk = item['two_1_x']
        bulk['alias_1__post'] = 'str-2'
        
        self.assertIn('alias_1', item)
        self.assertIn('alias_2', item)
        self.assertIn('two_1_1__alias_1__post', item)
        self.assertIn('two_1_1__alias_2__post', item)
        self.assertIn('two_1_1__alias_3__post', item)
        
        self.assertNotIn('wrong_key', item)
        self.assertNotIn('wrong_key__f_integer', item)
        self.assertNotIn('two_1_x__one_x_x__wrong_key', item)
        
        self.assertIn('alias_1__post', bulk)
        
        self.assertNotIn('wrong_key', bulk)
        self.assertNotIn('wrong_key__f_integer', bulk)
        self.assertNotIn('two_1_x__one_x_x__wrong_key', bulk)
        
        
        


from save_to_db.adapters.sqlalchemy_adapter import SqlachemyAdapter
from save_to_db.core.persister import Persister

from . import config
from .sqlalchemy_test_base import SqlalchemyTestBase

from .models.fields import ModelFieldTypes
from .models.constraints import (ModelConstraintsOne,
                                 ModelConstraintsTwo,
                                 ModelConstraintsThree,
                                 ModelConstraintsFour,
                                 ModelConstraintsFive,
                                 ModelConstraintsSix,
                                 ModelConstraintsSelf)
from .models.general import (ModelGeneralOne,
                             ModelGeneralTwo)
from .models.invalid_fields import ModelInvalidFieldNames
from .models.many_refs import ModelManyRefsOne, ModelManyRefsTwo

from .. import register_test
from ..test_fields_and_relations import TestFieldsAndRelations
from ..test_constraints import TestConstraints
from ..test_complete_item_metaclass import TestCompleteItemMetaclass
from ..test_item_cls_manager import TestItemClsManager
from ..test_item_fields import TestItemFields
from ..test_item_hooks import TestItemHooks
from ..test_item_persist import TestItemPersist
from ..test_item_dump import TestItemDump
from ..test_bulk_slicing import TestBulkSlicing
from ..test_scope import TestScope


AdapterPrefix = 'Sqlalchemy'
persister = Persister(SqlachemyAdapter,
                      adapter_settings={'session': config.session,
                                        'ModelBase': config.Base},
                      autocommit=True)
models = {
    'ModelFieldTypes': ModelFieldTypes,
    
    'ModelGeneralOne': ModelGeneralOne,
    'ModelGeneralTwo': ModelGeneralTwo,
    
    'ModelConstraintsOne': ModelConstraintsOne,
    'ModelConstraintsTwo': ModelConstraintsTwo,
    'ModelConstraintsThree': ModelConstraintsThree,
    'ModelConstraintsFour': ModelConstraintsFour,
    'ModelConstraintsFive': ModelConstraintsFive,
    'ModelConstraintsSix': ModelConstraintsSix,
    'ModelConstraintsSelf': ModelConstraintsSelf,
    
    'ModelInvalidFieldNames': ModelInvalidFieldNames,
    
    'ModelManyRefsOne': ModelManyRefsOne,
    'ModelManyRefsTwo': ModelManyRefsTwo,
}

for test_cls in (TestFieldsAndRelations,
                 TestConstraints,
                 TestItemClsManager,
                 TestCompleteItemMetaclass,
                 TestItemFields,
                 TestItemHooks,
                 TestItemPersist,
                 TestItemDump,
                 TestBulkSlicing,
                 TestScope):

    register_test(AdapterPrefix, SqlalchemyTestBase,
                  test_cls, models, persister)




from collections import defaultdict
from .merge_items import merge_items


def get_flat_item_structure(item):
    """ Takes an item, completes reverse relationships inside it and
    return a dictionary like this:
    
        .. code-block:: Python
            
            {
                item_cls: [item_instance, ...],
                ...
            }
            
    Where keys are a subclasses of :py:class:`~.item.Item` and values are
    instances of that class.
    """
    structure = defaultdict(list)  # item class to a list of instances
    __add_flat_item_to_structure(item, structure)
    
    # merging items
    all_items = []
    for item_cls in structure.keys():
        if not item_cls.allow_merge_items:
            continue
        if not all_items:
            for items in structure.values():
                all_items.extend(items)
        merge_items(structure[item_cls], all_items)
        
    return structure


def __add_flat_item_to_structure(item, structure):
    if item.is_single_item():
        if item in structure[type(item)]:
            return
        structure[type(item)].append(item)
        
        # setting reverse relations
        for key, relation in item.relations.items():
            if key not in item:
                continue
            
            that_item = item[key]
            reverse_key = relation['reverse_key']
            is_one_to_x = relation['relation_type'].is_one_to_x()

            if that_item.is_single_item():
                if is_one_to_x:
                    that_item[relation['reverse_key']] = item
                else:
                    that_item[relation['reverse_key']].add(item)
                __add_flat_item_to_structure(that_item, structure)
            else:
                for inner_item in that_item.bulk:
                    if is_one_to_x:
                        inner_item[reverse_key] = item
                    else:
                        inner_item[reverse_key].add(item)
                         
                    __add_flat_item_to_structure(inner_item, structure)
    else:
        # bulk item can be added (using this function) only once at the
        # beginning
        for inner_item in item.bulk:
            __add_flat_item_to_structure(inner_item, structure)


from save_to_db.core.exceptions.scope_except import ItemClsAlreadyScoped
from save_to_db.core.item import Item
from save_to_db.core.scope import Scope
from save_to_db.utils.test_base import TestBase



class TestScope(TestBase):
    
    ModelFieldTypes = None
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        
        class ItemFieldTypes(Item):
            model_cls = cls.ModelFieldTypes
        
        class ItemGeneralOne(Item):
            model_cls = cls.ModelGeneralOne
                
        class ItemGeneralTwo(Item):
            model_cls = cls.ModelGeneralTwo
        
        cls.ItemFieldTypes = ItemFieldTypes
        cls.ItemGeneralOne = ItemGeneralOne
        cls.ItemGeneralTwo = ItemGeneralTwo
    
    
    def test_relation_replacement(self):
        default_date_format = \
            self.ItemGeneralOne.conversions['date_format']
            
        new_date_format = '%d.%m.%Y'
        self.assertNotEqual(new_date_format, default_date_format)
        
        scope = Scope(fixes={
                self.ItemGeneralOne: {
                    'conversions': {
                        'date_format': new_date_format,
                    }
                }
            })
        self.assertEqual(len(scope._classes), 2)  # only 2 items scoped
        
        ScopedTypes, ScopedGeneralOne, ScopedGeneralTwo = \
            scope.get(self.ItemFieldTypes,
                      self.ItemGeneralOne,
                      self.ItemGeneralTwo)
        
        # no references to the scoped item, original returned
        self.assertIs(ScopedTypes, self.ItemFieldTypes)
        # scoped directly
        self.assertIsNot(ScopedGeneralOne, self.ItemGeneralOne)
        # scoped through reference
        self.assertIsNot(ScopedGeneralTwo, self.ItemGeneralTwo)
        
        # checking returned items
        expected = {
            ScopedTypes: default_date_format,
            ScopedGeneralOne: new_date_format,  # changed
            ScopedGeneralTwo: default_date_format,
            
            self.ItemFieldTypes: default_date_format,
            self.ItemGeneralOne: default_date_format,
            self.ItemGeneralTwo: default_date_format,
        }
        for item_cls, expected_date_format in expected.items():
            actual_date_format = item_cls.conversions['date_format']
            self.assertEqual(actual_date_format, expected_date_format)

        # relations of scoped items fixed
        two_1_1_cls = ScopedGeneralOne.relations['two_1_1']['item_cls']
        self.assertIs(two_1_1_cls, ScopedGeneralTwo)
        
        one_x_x_cls = ScopedGeneralTwo.relations['one_x_x']['item_cls']
        self.assertIs(one_x_x_cls, ScopedGeneralOne)
        
        # relations of not scoped items left as they were
        two_1_1_cls = \
            self.ItemGeneralOne.relations['two_1_1']['item_cls']
        self.assertIs(two_1_1_cls, self.ItemGeneralTwo)
        
        one_x_x_cls = \
            self.ItemGeneralTwo.relations['one_x_x']['item_cls']
        self.assertIs(one_x_x_cls, self.ItemGeneralOne)
    
    
    def test_default_item_cls(self):
        default_date_format = \
            self.ItemGeneralOne.conversions['date_format']
            
        new_date_format = '%d.%m.%Y'
        self.assertNotEqual(new_date_format, default_date_format)
        
        new_default_date_format = '%m/%d/%Y'
        self.assertNotEqual(new_default_date_format, default_date_format)
        
        scope = Scope(fixes={
                self.ItemGeneralOne: {
                    'conversions': {
                        'date_format': new_date_format,
                    }
                },
                None: {
                    'conversions': {
                        'date_format': new_default_date_format,
                    }
                }
            })
        self.assertEqual(len(scope._classes), 3)  # all items scoped
        
        ScopedTypes, ScopedGeneralOne, ScopedGeneralTwo = \
            scope.get(self.ItemFieldTypes,
                      self.ItemGeneralOne,
                      self.ItemGeneralTwo)
        
        # checking returned items
        expected = {
            # scoped
            ScopedTypes: new_default_date_format,
            ScopedGeneralOne: new_date_format,  # changed directly
            ScopedGeneralTwo: new_default_date_format,
            
            self.ItemFieldTypes: default_date_format,
            self.ItemGeneralOne: default_date_format,
            self.ItemGeneralTwo: default_date_format,
        }
        for item_cls, expected_date_format in expected.items():
            actual_date_format = item_cls.conversions['date_format']
            self.assertEqual(actual_date_format, expected_date_format)
            
    
    def test_item_already_scoped_exception(self):
        scope = Scope(fixes={
                self.ItemFieldTypes: {
                    'conversions': {
                        'date_format': '%m/%d/%Y',
                    }
                },
            })
        ScopedTypes = scope.get(self.ItemFieldTypes)[0]
        
        with self.assertRaises(ItemClsAlreadyScoped):
            scope = Scope(fixes={
                    ScopedTypes: {
                        'conversions': {
                            'date_format': '%d-%m-%Y',
                        }
                    },
                })
    
    
    def test_item_cls_autogeneration(self):
        self.item_cls_manager.clear()
        self.item_cls_manager.autogenerate = True
        
        default_date_format = \
            self.ItemGeneralOne.conversions['date_format']
        new_date_format = '%m/%d/%Y'
        self.assertNotEqual(new_date_format, default_date_format)
        
        class ItemFieldTypes(Item):
            model_cls = self.ModelFieldTypes
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
        
        scope = Scope(fixes={
                ItemGeneralOne: {
                    'conversions': {
                        'date_format': new_date_format,
                    }
                },
            })
        
        # 2 items including related (ItemFieldTypes discarded)
        self.assertEqual(len(scope._classes), 2)
        
        ItemGeneralTwo = ItemGeneralOne.relations['two_x_x']['item_cls']
        
        self.assertTrue(ItemGeneralTwo.model_cls, self.ModelGeneralTwo)
        
        self.assertIn(ItemGeneralOne, scope._classes)
        self.assertIn(ItemGeneralTwo, scope._classes)
        
        self.assertFalse(ItemFieldTypes.metadata['autogenerated_item_cls'])
        self.assertFalse(ItemGeneralOne.metadata['autogenerated_item_cls'])
        self.assertTrue(ItemGeneralTwo.metadata['autogenerated_item_cls'])
        
        self.assertEqual(ItemFieldTypes.conversions['date_format'],
                         default_date_format)
        self.assertEqual(ItemGeneralOne.conversions['date_format'],
                         default_date_format)
        self.assertEqual(ItemGeneralTwo.conversions['date_format'],
                         default_date_format)
        
        
        self.assertEqual(scope[ItemFieldTypes].conversions['date_format'],
                         default_date_format)
        self.assertEqual(scope[ItemGeneralOne].conversions['date_format'],
                         new_date_format)
        self.assertEqual(scope[ItemGeneralTwo].conversions['date_format'],
                         default_date_format)
        
    
    def test_item_cls_autogeneration_and_default(self):
        self.item_cls_manager.clear()
        self.item_cls_manager.autogenerate = True
        
        default_date_format = \
            self.ItemGeneralOne.conversions['date_format']
            
        new_date_format = '%m/%d/%Y'
        self.assertNotEqual(new_date_format, default_date_format)
        
        new_default_date_format = '%m/%d/%Y'
        self.assertNotEqual(new_default_date_format, default_date_format)
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
        
        scope = Scope(fixes={
                ItemGeneralOne: {
                    'conversions': {
                        'date_format': new_date_format,
                    }
                },
                None: {
                    'conversions': {
                        'date_format': new_default_date_format,
                    }
                }
            })
        
        # 1 related item was also added
        self.assertEqual(len(scope._classes), 2)
        
        ItemGeneralTwo = ItemGeneralOne.relations['two_x_x']['item_cls']
        
        self.assertTrue(ItemGeneralTwo.model_cls, self.ModelGeneralTwo)
        
        self.assertIn(ItemGeneralOne, scope._classes)
        self.assertIn(ItemGeneralTwo, scope._classes)
        
        self.assertFalse(ItemGeneralOne.metadata['autogenerated_item_cls'])
        self.assertTrue(ItemGeneralTwo.metadata['autogenerated_item_cls'])
        
        self.assertEqual(ItemGeneralOne.conversions['date_format'],
                         default_date_format)
        self.assertEqual(ItemGeneralTwo.conversions['date_format'],
                         default_date_format)
        
        self.assertEqual(scope[ItemGeneralOne].conversions['date_format'],
                         new_date_format)
        self.assertEqual(scope[ItemGeneralTwo].conversions['date_format'],
                         new_default_date_format)
        
        
        
        


import json
from pprint import pprint
from .utils.dict_wrapper import DictWrapper


class ItemBase(object):
    """ This is an abstract item class that serves as a base class for
    single item and bulk item classes.
    
    .. seealso::
        :py:class:`~.item.Item` and :py:class:`~.bulk_item.BulkItem` classes.
    
    :param \*\*kwargs: Values that will be saved as item data.
    """
    
    #--- special methods -------------------------------------------------------
    
    def __init__(self, **kwargs):
        self.complete_setup()

        self.data = {}
        for key, value in kwargs.items():
            item = self
            real_keys = self._get_real_keys(key)
            i = -1
            for i, real_key in enumerate(real_keys[:-1]):
                if item.is_bulk_item():
                    break
                item = item[real_key]
            
            remaining_key = '__'.join(real_keys[i+1:])
            if not isinstance(value, list):
                item[remaining_key] = value
            else:
                item[remaining_key].add(*value)
                
    
    def __setitem__(self, key, value):
        raise NotImplementedError()
    
    
    def __call__(self, **kwargs):
        for key, value in kwargs.items():
            self[key] = value
        return self
        
    
    def __getitem__(self, key):
        raise NotImplementedError()
    
    
    def __delitem__(self, key):
        raise NotImplementedError()
    
    
    def __contains__(self, key):
        raise NotImplementedError()
    
    
    #--- utility methods -------------------------------------------------------
         
    def to_dict(self):
        """ Converts item into a Python `dict` object.

        :returns: Python `dict` representation of the item.
        """
        raise NotImplementedError()
    
    
    def dict_wrapper(self):
        """ This method is used for integration with Scrapy project, when
        parsing pages in Scrapy you can yield an item as
        :py:class:`~.utils.dict_wrapper.DictWrapper` (subclass of `dict`)
        and then use :py:meth:`~.utils.dict_wrapper.DictWrapper.get_item`
        method to get the original item.
        
        :returns: An :py:class:`~.utils.dict_wrapper.DictWrapper` class
            instance.
        """
        return DictWrapper(self)
    
    
    def load_dict(self, data):
        """ Loads data from dictionary into the item.
        
        :param data: Dictionary with item data (see :func:`~to_dict` method).
        :returns: The item itself.
        """
        raise NotImplementedError()
    
    
    #--- helper functions ------------------------------------------------------
    
    def pprint(self, as_json=True):
        """ Pretty prints the item. """
        if not as_json:
            pprint(self.to_dict())
        else:
            print(json.dumps(self.to_dict(), indent=4, sort_keys=True))
    
    
    #--- main methods ----------------------------------------------------------
    
    def as_list(self):
        """ Returns a list of items. For single items simply returns a list 
        containing only `self`, for bulk items returns list of all items.  
        """
        raise NotImplementedError()
    
    
    def is_single_item(self):
        """ Returns `True` if item is a single item. """
        raise NotImplementedError()
    
    
    def is_bulk_item(self):
        """ Returns `True` if item is a bulk item. """
        raise NotImplementedError()
    
    
    def process(self):
        """ Converts all set string values to the appropriate data types and
        sets default values if needed.
        """
        raise NotImplementedError()
    


import pprint


class DictWrapper(dict):
    """ Used as a temporary wrapper around
    :py:class:`~save_to_db.core.item_base.ItemBase` instance for compatibility
    with Scrapy project.
    
    Scrapy is an open source and collaborative framework for extracting the data
    you need from websites. In a fast, simple, yet extensible way.
    [#DictWrapper_1]_

    When parsing a page with Scrapy you cannot yield instances of arbitrary
    classes, but you can yield an instance of a `dict` class which is treated
    as an instance of a Scrapy item. Using
    :py:meth:`~save_to_db.core.item_base.ItemBase.to_dict` and then
    :py:meth:`~save_to_db.core.item_base.ItemBase.load_dict` of
    :py:class:`~save_to_db.core.item_base.ItemBase` instance is expensive, as it
    properly transforms an item into a `dict` and then `dict` to item. Using
    :py:meth:`~save_to_db.core.item_base.ItemBase.dict_wrapper` method of
    :py:class:`~save_to_db.core.item_base.ItemBase` will just wrap an item in
    an `DictWrapper` instance (subclass of `dict` class),
    later you can get the wrapped item from the wrapper in a Scrapy pipeline. 
    
    .. [#DictWrapper_1] Citation from https://scrapy.org site.
    
    :param item: An instance of :py:class:`~save_to_db.core.item_base.ItemBase`
        to be wrapped in a dictionary.
    """
    
    def __init__(self, item):
        self['item'] = item
    
    def get_item(self):
        """ Returns an originally wrapped item instance.
        
        :returns: Instance of
            :py:class:`~save_to_db.core.item_base.ItemBase` class.
        """
        return self['item']
    
    def __str__(self):
        return pprint.pformat(self.get_item().to_dict())


import copy
import uuid
from .exceptions.scope_except import ItemClsAlreadyScoped
from .item_cls_manager import item_cls_manager



class Scope(object):
    """ Class for scoping :py:class:`~.item.Item` classes.
    
    :param fixes: a dictionary whose keys are item classes (or `None`
        for all classes not in the dictionary) and values are class
        attributes to be replaced.
    
    Example usage:
        
        .. code:: Python
        
            class TestItem(Item):
                model_cls = SomeModel
            
            scope = Scope({
                TestItem: {
                    'conversions': {
                        'date_format: '%m/%d/%Y',
                    },
                },
                None: {
                    'conversions': {
                        'date_format: '%d.%m.%Y',
                    },
                },
            })
            
            ScopedTestItem = scope[TestItem]  # or `scope.get(TestItem)`
    
    When an item is scoped other items that use the original item in
    relations are also scoped and their relation data fixed.
    """
    
    def __init__(self, fixes):
        self.scope_id = uuid.uuid4().int
        self._classes = {}
        item_cls_manager.autocomplete_item_classes()
        
        # fixing classes
        for item_cls, item_fixes in fixes.items():
            if item_cls is None or not item_fixes:
                continue
            self.__prepare_scoped_item_cls(item_cls, item_fixes)
        
        # default
        if None in fixes and fixes[None]:
            for item_cls in item_cls_manager._registry:
                if item_cls in self._classes:
                    continue
                self.__prepare_scoped_item_cls(item_cls, fixes[None])
        
        # fixing relations
        do_continue = True
        while do_continue:
            do_continue = False
            #--- first scoping items that need to be scoped ---
            for item_cls in item_cls_manager._registry:
                if item_cls in self._classes:
                    continue
                
                item_cls.complete_setup()
                for rel_data in item_cls.relations.values():
                    if rel_data['item_cls'] in self._classes:
                        self.__prepare_scoped_item_cls(item_cls, item_fixes={})
                        break
            
            #--- updating relations if needed ---
            for scoped_item_cls in self._classes.values():
                relations = scoped_item_cls.relations
                for key, rel_data in relations.items():
                    rel_item_cls = rel_data['item_cls']
                    if rel_item_cls not in self._classes:
                        continue
                    
                    relations[key]['item_cls'] = self._classes[rel_item_cls]
                    do_continue = True
    
    
    def __prepare_scoped_item_cls(self, item_cls, item_fixes):
        if item_cls.metadata['scope_id'] != 0:
            raise ItemClsAlreadyScoped(item_cls)
        
        item_cls.complete_setup()
        item_fixes['relations'] = copy.deepcopy(item_cls.relations)
        item_fixes['metadata'] = copy.deepcopy(item_cls.metadata)
        item_fixes['metadata']['scope_id'] = self.scope_id

        class_name = item_cls.__name__
        scoped_item_cls = type('Scoped{}'.format(class_name), (item_cls,),
                               item_fixes)
        scoped_item_cls.complete_setup()
        self._classes[item_cls] = scoped_item_cls
    
    
    def __getitem__(self, item_cls):
        if item_cls in self._classes:
            return self._classes[item_cls]
        else:
            return item_cls
    
    
    def get(self, *item_classes):
        """ Excepts non-scoped items and returns corresponding scoped items or
        original items for items not present in the scope.
        
        :param item_classes: Items for which scoped item versions going to be
            returned.
        :returns: List of scoped items. If scope does not have corresponding
            scoped version of an item class, original class is used.
        """
        return [self[item_cls] for item_cls in item_classes]



from collections import defaultdict
from save_to_db.adapters.utils.relation_type import RelationType
from save_to_db.core.exceptions import MultipleModelsMatch
from save_to_db.core.item_base import ItemBase
from .flat_item_structure import get_flat_item_structure
from .match_items_to_models import match_items_to_models


def db_persist(adapter, item, adapter_settings):
    item.process()
    result_items = item.as_list()
    
    # creating process keeper
    process_keeper = defaultdict(list)
    for item_cls, items in get_flat_item_structure(item).items():
        for item in items:
            process_keeper[item_cls].append(__ItemProcessTracker(item))
    
    def keeper_is_empty(process_keeper):
        for value in process_keeper.values():
            if value:
                return False
        return True
    
    # --- getting and updating initial models ---
    getter_process_keeper = process_keeper
    while not keeper_is_empty(getter_process_keeper):
        all_items, all_models = [], []
        for item_cls in getter_process_keeper:
            if not getter_process_keeper[item_cls]:
                continue
            
            items_and_fkeys = []
            new_items, new_models = [], []
            
            for item_track in getter_process_keeper[item_cls]:
                items_and_fkeys.append([item_track.item,
                                        item_track.fkeys])
            
            got_models = adapter.get(items_and_fkeys, adapter_settings)
            got_items, got_models = match_items_to_models(adapter,
                                                          items_and_fkeys,
                                                          got_models)
            all_items.extend(got_items)
            all_models.extend(got_models)

        getter_process_keeper = __update_relationships(
            adapter, adapter_settings, all_items, all_models, process_keeper)[1]
    
    
    for item_cls in process_keeper:
        for item_track in process_keeper[item_cls]:
            item = item_track.item
            models = item_track.models
            for model in models:
                item.before_model_update(model)  # hook
                
                # cleaning x-to-many relations if needed
                for fkey in item_cls.relations:
                    relation = item_cls.relations[fkey]
                    if not relation['relation_type'].is_x_to_many() or \
                            not relation['replace_x_to_many'] or \
                            fkey not in item:
                        continue
                    __clear_related_models(adapter, adapter_settings,
                                           relation,
                                           model, fkey, process_keeper)
                
                __update_model_fields(adapter, adapter_settings,
                                      item_track, model, process_keeper)
            
    # --- creation loop ---
    create_process_keeper = process_keeper
    while not keeper_is_empty(create_process_keeper):
        new_items, new_models = [], []
        for item_cls in create_process_keeper:
            for item_track in create_process_keeper[item_cls]:
                if item_track.models:
                    continue
                item = item_track.item
                fkeys = item_track.fkeys
                if item.can_create(fkeys):
                    new_items.append(item)
                    model = adapter.create_blank_model(item.model_cls,
                                                       adapter_settings)
                    new_models.append([model])
                    item.before_model_update(model)  # hook
                    __update_model_fields(adapter, adapter_settings,
                                          item_track, model, process_keeper,
                                          save=True)

        create_process_keeper = __update_relationships(
            adapter, adapter_settings, new_items, new_models, process_keeper)[0]
            
    # after_model_update hook
    for item_cls in process_keeper:
        for item_track in process_keeper[item_cls]:
            item = item_track.item
            for model in item_track.models:
                if adapter.SAVE_MODEL_BEFORE_COMMIT:
                    adapter.save_model(model, adapter_settings)
                item.after_model_update(model)
    
    # return the result
    items, models = [], []
    for item in result_items:
        # looking for models
        for item_track in process_keeper[type(item)]:
            if item is item_track.item:
                if item_track.models:
                    items.append(item_track.item)
                    models.append(item_track.models)
                break
        
    return items, models


# functions for model updates --------------------------------------------------

def __clear_related_models(adapter, adapter_settings, relation,
                           model, fkey, process_keeper):
    if not adapter.REVERSE_MODEL_AUTOUPDATE_SUPPORTED and \
            relation['reverse_key'] and relation['item_cls'] and \
            relation['relation_type'] is RelationType.ONE_TO_MANY:
        item_cls = relation['item_cls']
        for related_model in adapter.get_related_x_to_many(model, fkey):
            for item_track in process_keeper[item_cls]:
                for track_model in item_track.models:
                    if track_model == related_model:
                        __set_related_model(adapter, adapter_settings,
                                            process_keeper, item_track,
                                            track_model,
                                            relation['reverse_key'], None)
        
    adapter.clear_related_models(model, fkey)


def __set_related_model(adapter, adapter_settings, process_keeper, item_track,
                        model, fkey, fmodel):
    
    def clean_relation(model, fkey, fmodel, item_cls, reverse_key):
        if reverse_key and hasattr(model, fkey):
            related_model = getattr(model, fkey)
            if related_model and related_model != fmodel:
                setattr(related_model, reverse_key, None)
                adapter.save_model(related_model, adapter_settings)
         
        for item_track in process_keeper[item_cls]:
            found = False
            for track_model in item_track.models:
                if track_model == model:
                    continue
                if hasattr(track_model, fkey) and \
                        getattr(track_model, fkey) == fmodel:
                    found = True
                    setattr(track_model, fkey, None)
                    adapter.save_model(track_model, adapter_settings)
                    break
             
            if found:
                break
             
    relation = item_track.item.relations[fkey]
    if relation['relation_type'] is RelationType.ONE_TO_ONE and \
            ((not hasattr(model, fkey) and fmodel is not None) or
             (hasattr(model, fkey) and getattr(model, fkey) != fmodel)):
        # forward relation
        clean_relation(model, fkey, fmodel, type(item_track.item),
                       relation['reverse_key'])
        # backward
        if relation['reverse_key']:
            clean_relation(fmodel, relation['reverse_key'], model,
                           relation['item_cls'],
                           fkey)
                
    setattr(model, fkey, fmodel)
    
    

def __add_related_models(adapter, model, fkey, fmodels):
    adapter.add_related_models(model, fkey, fmodels)

# process tracker --------------------------------------------------------------
    
class __ItemProcessTracker(object):
    def __init__(self, item):
        self.item = item
        self.models = []
        self.fkeys = defaultdict(list)


def __update_model_fields(adapter, adapter_settings, item_track,
                          model, process_keeper, save=False):
    for field_name in item_track.item.data:
        if field_name in item_track.item.fields:
            setattr(model, field_name, item_track.item[field_name])
        
    # many to one before saving
    for fkey in item_track.fkeys:
        relation_type = item_track.item.relations[fkey]['relation_type']
        if relation_type.is_x_to_one():
            __set_related_model(adapter, adapter_settings,
                                process_keeper, item_track,
                                model, fkey, item_track.fkeys[fkey][0])
    
    # for the rest model has to be saved
    if save and adapter.SAVE_MODEL_BEFORE_COMMIT:
        adapter.save_model(model, adapter_settings)

    for fkey, fmodels in item_track.fkeys.items():
        relation_type = item_track.item.relations[fkey]['relation_type']
        if relation_type.is_x_to_one():
            continue
        if relation_type.is_x_to_one():
            __set_related_model(adapter, adapter_settings,
                                process_keeper, item_track,
                                model, fkey, fmodels[0])
        else:
            __add_related_models(adapter, model, fkey, fmodels)

def __update_relationships(adapter, adapter_settings,
                           items, models, process_keeper):
    original_items, original_models = items, models
    
    creator_keeper = defaultdict(list)
    getter_keeper = defaultdict(list)
    
    # update_fkey --------------------------------------------------------------
    
    def update_fkey(item_track, fkey, foreign_item_track):

        def update_result_keepers():
            item_cls = type(item)
            for keeper, groups in [[creator_keeper[item_cls], item.creators],
                                   [getter_keeper[item_cls], item.getters]]:
                if groups is None:
                    continue
                
                for group in groups:
                    if fkey not in group:
                        continue
                    
                    do_add = True
                    for field in group:
                        if field not in item:
                            do_add = False
                            break
                    if do_add:
                        keeper.append(item_track)
                        break
        
        foreign_models = foreign_item_track.models
        if not foreign_models:
            return
        
        item = item_track.item
        models = item_track.models
        fkeys = item_track.fkeys[fkey]
        
        relation = item.relations[fkey]
        is_x_to_one = relation['relation_type'].is_x_to_one()
        
        keepers_updated = False
        for foreign_model in foreign_models:
            if foreign_model not in fkeys:
                fkeys.append(foreign_model)
                if not keepers_updated:
                    update_result_keepers()
        
#         print('     REF:', item.__class__.__name__, '[{}]->'.format(fkey),
#               foreign_item_track.item.__class__.__name__,
#               '(is_x_to_one: {})'.format(is_x_to_one))
        
        for model in models:
            if is_x_to_one:
                if len(foreign_models) > 1:
                    raise MultipleModelsMatch(item, fkey, foreign_models)
                __set_related_model(adapter, adapter_settings,
                                    process_keeper, item_track,
                                    model, fkey, foreign_models[0])
            else:
                __add_related_models(adapter, model, fkey, foreign_models)
    
    
    # saving models to process_keeper ------------------------------------------
    
    for item, models in zip(original_items, original_models):
        if not models:
            continue
        item_cls = type(item)
        
        # getting item_track
        for item_track in process_keeper[item_cls]:
            if item_track.item is item:
                break
            
        for model in models:
            if model not in item_track.models:
                item_track.models.append(model)
                
                if len(item_track.models) > 1 and \
                        not item_cls.allow_multi_update:
                    raise MultipleModelsMatch(
                        item, item_track.models)
            
    # updating relationships ---------------------------------------------------
    
    for item, models in zip(original_items, original_models):
        if not models:
            continue
        item_cls = type(item)
        
        # getting item_track
        for item_track in process_keeper[item_cls]:
            if item_track.item is item:
                break
               
        # --- forward relations ---
        for f_fkey in item_cls.relations:
            if f_fkey not in item:
                continue
            
            for f_item in item[f_fkey].as_list():
                
                # looking for f_item_track
                found = False
                for f_item_cls in process_keeper:
                    for f_item_track in process_keeper[f_item_cls]:
                        if f_item_track.item is f_item:
                            update_fkey(item_track, f_fkey, f_item_track)
                            found = True
                            break
                    if found:
                        break

                # updating f_fkey
                
            
        # --- backward relations ---
        for b_item_cls in process_keeper:
            # looking for b_item_track
            for b_item_track in process_keeper[b_item_cls]:
                b_item = b_item_track.item

                for b_fkey, b_value in b_item.data.items():
                    # not an item?
                    if not isinstance(b_value, ItemBase):
                        continue
                     
                    # not referenced?
                    if b_value.is_single_item():
                        if not (b_value is item):
                            continue
                    else:
                        if item not in b_value.bulk:
                            continue
                    
                    # updating b_fkey
                    update_fkey(b_item_track, b_fkey, item_track)
    
    return creator_keeper, getter_keeper


""" Module that contains tests for adapters. The tests require next models:

    Models for basic adapter tests:

        - `ModelFieldTypes` with columns defined according to the rules:
            
            - "id" can be present or not (ignored).
            - Model fields (table columns) must be prefixed with it's type and
              underscore, for example: "integer_first". Types are just lower
              case versions of
              :py:class:`~save_to_db.adapters.utils.column_type.ColumnType`
              enumeration members.
            - Column names in database must be prefixed with 'new_'.
            - All types must be used at least once, except for `OTHER` type
              which must be not used at all.
    
    Models for basic relationship tests, must contains all types of
    relationships between different models and with self:
    
        - `ModelGeneralOne` and `ModelGeneralTwo`
          with next fields, relations and other definitions:
          
            - `ModelGeneralOne` must have "model_general_one" table name,
              `ModelGeneralTwo` must have "model_general_two" table name.
            - "id" primary key and any other fields used only as foreign keys.
            - both table must have next self referencing fields: 
                
                - "id" primary key
                - `parent_x_1` many-to-one relationship, `child_1_x` many-to-one
                  relationship on the other side.
                - `parent_1_1` one-to-one relationship, `child_1_1` one-to-one
                  relationship on the other side.
                - `parent_x_x` many-to-many relationship, `child_x_x`
                  many-to-many relationship on the other side.
                - "f_binary", "f_boolean", "f_string", "f_text", "f_integer",
                  "f_float", "f_date", "f_time", "f_datetime" columns with
                  appropriate types
            
            - `ModelGeneralOne.two_1_x` one-to-many relationship,
              `ModelGeneralTwo.one_x_1` many-to-one relationship
              on the other side.
            - `ModelGeneralOne.two_1_1` one-to-one relationship,
              `ModelGeneralTwo.one_1_1` one-to-one relationship
              on the other side.
            - `ModelGeneralOne.two_x_x` many-to-many
              relationship, `ModelGeneralTwo.one_x_x` many-to-many
              relationship on the other side.
            - `ModelGeneralOne.two_x_1` many-to-one relationship,
              `ModelGeneralTwo.one_1_x` one-to-many relationship
              on the other side.
              
            - all foreign keys must be "delete on cascade".
            
        - `ModelManyRefsOne` and `ModelManyRefsTwo`
          with next fields, relations and other definitions:
          
            - `ModelManyRefsOne` must have "model_many_refs_one" table name,
              `ModelManyRefsTwo` must have "model_many_refs_two" table name.
            - "id" primary key.
            
            - `ModelManyRefsOne.two_x_1_a` many-to-one relationship,
              `ModelManyRefsTwo.one_1_x_a` one-to-many relationship
              on the other side.
            - `ModelManyRefsOne.two_1_x_b` one-to-many relationship,
              `ModelManyRefsTwo.one_x_1_b` many-to-many relationship
              on the other side.
              
            - `ModelManyRefsOne.two_x_x_a` many-to-many relationship,
              `ModelManyRefsTwo.one_x_x_a` many-to-many relationship
              on the other side.
            - `ModelManyRefsOne.two_x_x_b` many-to-many relationship,
              `ModelManyRefsTwo.one_x_x_b` many-to-many relationship
              on the other side.
              
            - `ModelManyRefsOne.two_1_1_a` one-to-one relationship,
              `ModelManyRefsTwo.one_1_1_a` one-to-one relationship
              on the other side. Table one has foreign key to table two.
            - `ModelManyRefsOne.two_1_1_b` one-to-one relationship,
              `ModelManyRefsTwo.one_1_1_b` one-to-one relationship
              on the other side. Table two has foreign key to table one.
              
            - all foreign keys must be "delete on cascade".
    
    Models for creators and getters auto configuration should contain:
    
        - Models with single primary key and relationships using it:
        
            - many-to-one with required value.
            - many-to-one with not required value.
            
        - Models with composite primary key:
        
            - many-to-one with required value.
            - many-to-one with not required value.
            
        - Models null constraint:
            
            - Model with not null field.
        
        - Models with unique constraints:
            
            - Model with simple unique field.
            - Model with simple unique field that cannot be null.
            - Model with composite unique constraint.
            - Model with simple foreign key that has a unique constraint.
            - Model with simple not null foreign key that has a unique
              constraint.
            - Model with simple foreign key that has a unique constraint
              together with other column.
            - Model with composite foreign key that has a unique constraint.
        
        - Model with many-to-many-relationships.
        - Model with many-to-one not null relation with itself, one-to-one
          relationship with unique constraint with itself, one-to-one not null
          relationship with unique constraint with itself, many-to-many
          relationship with itself.
        - All models must have "f_binary", "f_boolean", "f_string", "f_text",
          "f_integer", "f_float", "f_date", "f_time", "f_datetime" columns with
          appropriate types.
        
    And the models are:
    
        - `ModelConstraintsOne`:
            
            - `model_constraints_one` table name
            - `id` primary key
            - `f_text` cannot be null
            - `f_integer` unique
            - `f_string` not null and unique
            
        - `ModelConstraintsTwo`:
        
            - `model_constraints_two` table name
            - `id` primary key
            - `one_x_1_id` foreign key to `ModelConstraintsOne` with `one_x_1`
              relation and `two_1_x` on the other side
            - `three_x_x` relation with `ModelConstraintsThree` with `two_x_x`
              on the other side
            - `four_primary_one` and `four_primary_two` both not null foreign
              keys to `ModelConstraintsFour` with `four_x_1` relation and
              `two_1_x` on the other side
        
        - `ModelConstraintsThree`:
        
            - `model_constraints_three` table name
            - `id` primary key
            - `one_x_1_id` not null foreign key to `ModelConstraintsOne` with
              `one_x_1` relation and `three_1_x` on the other side
            - `two_x_x` relation with `ModelConstraintsTwo` and `three_x_x` on
              the other side
            - `four_primary_one` and `four_primary_two` foreign keys to
              `ModelConstraintsFour` that can be null, `four_x_1` relation with
              `three_1_x` on the other side
        
        - `ModelConstraintsFour`:
        
            - `model_constraints_four` table name
            - composite `primary_one` plus `primary_two` columns primary key
            - `f_integer` plus `f_string` unique constraint
            
        - `ModelConstraintsFive`:
        
            - `model_constraints_five` table name
            - `id` primary key
            - `one_1_1_id` unique foreign key to `ModelConstraintsOne` with
              `one_1_1` relation and `five_1_1` on the other side
            - `four_primary_one` plus `four_primary_two` unique foreign key
              to `ModelConstraintsFour` with `four_1_1` relation and `five_1_1`
              on the other side
        
        - `ModelConstraintsSix`:
        
            - `model_constraints_six` table name
            - `id` primary key
            - `five_1_1_id` not null foreign key to `ModelConstraintsFive` with
              `five_1_1` relation and `six_1_1` on the other side
            - `five_1_1_id` plus `f_integer` unique constraint
            
        - `ModelConstraintsSelf`:
        
            - `model_constraints_self` table name
            - `id` primary key
            - `parent_x_1_id` not null foreign key to itself with `parent_x_1`
              relation and `child_1_x` on the other side.
            - `first_parent_1_1_id` foreign key with unique constraint to
              itself with `first_parent_1_1` relation and `first_child_1_1` on
              the other side.
            - `parent_x_x` many-to-many relationship with itself with
              `child_x_x` on the other side.
    
    Model for testing default model values:
    
        - `ModelDefaultValues`:
        
            - `model_default_values` table name.
            - `id` primary key.
            - `f_integer` with no default value, not null.
            - `f_boolean` with `True` default value, not null.
            - `f_string` with 'default_string' default value, not null.
            - `f_text` with default value defined by a function, not null.
"""

import importlib
from unittest import TestCase


# tuple of database adapter names to test
ADAPTER_NAMES = (
    'sqlalchemy_adapter',
    'django_adapter',
)


def init():
    """ Loads tests for all adapters. (Loading `tests` module from
    an adapter test module must trigger registration of tests)
    """
    for adapter_name in ADAPTER_NAMES:
        module_name = \
            'save_to_db.tests.test_adapters.test_{}.register_tests'.format(
                adapter_name)
        importlib.import_module(module_name)
        

registered_tests = []

def register_test(adapter_prefix, test_base, test_case, models,
                  persister):
    """ Registers a test case to be run on test command.
    
    :param adapter_prefix: A string prefix that will be used as part of the
        test class name.
    :param test_base: Test base class to be added as one of the base classes of
        a new test case. Must contain code to create and drop tables
        from database.
    :param test_case: The test case with tests.
    :param models: A dictionary of models that are needed for tests to run.
    :param persister: An instance of
        :py:class:`~save_to_db.core.persistor.Persistor` (`autocommit`
        must be `True`).
    """
    global registered_tests
    
    new_test_case_dict = {
        'persister': persister,
        'adapter_prefix': adapter_prefix,
    }
    new_test_case_dict.update(models)
    new_test_case = type('{}{}'.format(adapter_prefix, test_case.__name__),
                         (test_base, test_case),
                         new_test_case_dict)
    registered_tests.append(new_test_case)



from django.apps import apps

from django.db import models, transaction
from django.db.models import Q

from django.db.models.fields.related import OneToOneField, OneToOneRel
from django.db.models.fields.related import ManyToManyField, ManyToManyRel
from django.db.models.fields.related import ManyToOneRel
from django.db.models.fields.related import ForeignKey

# from django.db.models.fields.related_descriptors import \
#     ReverseOneToOneDescriptor, ReverseManyToOneDescriptor

from .utils.adapter_base import AdapterBase
from .utils.column_type import ColumnType
from .utils.relation_type import RelationType



class DjangoAdapter(AdapterBase):
    """ An adapter for working with Django ORM.
    
    The `adapter_settings` is a dictionary with next values:
    
        - *using* is a database name (a key from `DATABASES` dictionary in the
          "settings.py" file of a Django project), if it isn't provided
          "default" database is used.
    
    .. note::
    
        If you are going to use transactions, set `autocommit` for the database
        to `False`:
        
            .. code-block:: Python
        
                from django.db import transaction
                # Next function takes a `using` argument which should be the
                # name of a database. If it isn’t provided, Django uses the
                # "default" database.
                transaction.set_autocommit(False, using='database_name')
    """
    
    COMPOSITE_KEYS_SUPPORTED = False
    REVERSE_MODEL_AUTOUPDATE_SUPPORTED = False
    SAVE_MODEL_BEFORE_COMMIT = True
    
    
    #--- general methods -------------------------------------------------------
    
    @classmethod
    def is_usable(cls, model_cls):
        return issubclass(model_cls, models.Model)
    
    
    @classmethod
    def commit(cls, adapter_settings):
        transaction.commit(using=adapter_settings.get('using', 'default'))


    @classmethod
    def rollback(cls, adapter_settings):
        transaction.rollback(using=adapter_settings.get('using', 'default'))
    
    
    @classmethod
    def iter_fields(cls, model_cls):
        foreign_key_classes = (ForeignKey, ManyToOneRel,
                               OneToOneField, OneToOneRel,
                               ManyToManyField, ManyToManyRel)
        
        type_to_const = (
            (models.BinaryField, ColumnType.BINARY),
            (models.TextField, ColumnType.TEXT),
            (models.CharField, ColumnType.STRING),
            ((models.AutoField, models.IntegerField), ColumnType.INTEGER),
            ((models.BooleanField,
              models.NullBooleanField), ColumnType.BOOLEAN),
            (models.DateTimeField, ColumnType.DATETIME),
            ((models.FloatField,
              models.DecimalField), ColumnType.FLOAT),
            (models.DateField, ColumnType.DATE),
            (models.TimeField, ColumnType.TIME),
            (models.DateTimeField, ColumnType.DATETIME),
        )
        for field in model_cls._meta.get_fields():
            if isinstance(field, foreign_key_classes):
                continue
            
            yield_type = ColumnType.OTHER
            for field_class, column_type in type_to_const:
                if isinstance(field, field_class):
                    yield_type = column_type
                    break
            yield field.name, yield_type
    
    
    @classmethod
    def iter_relations(cls, model_cls):
        foreign_key_classes = (ForeignKey, ManyToOneRel,
                               OneToOneField, OneToOneRel,
                               ManyToManyField, ManyToManyRel)
        
        type_to_const = [
            [(OneToOneField, OneToOneRel), RelationType.ONE_TO_ONE],
            [ForeignKey, RelationType.MANY_TO_ONE],
            [ManyToOneRel, RelationType.ONE_TO_MANY],
            [(ManyToManyField,
              ManyToManyRel), RelationType.MANY_TO_MANY],
        ]
        for field in model_cls._meta.get_fields():
            if not isinstance(field, foreign_key_classes):
                continue
            
            for field_class, column_type in type_to_const:
                if isinstance(field, field_class):
                    yield_type = column_type
                    break
            
            yield field.name, field.related_model, yield_type, \
                field.remote_field.name
    
    
    @classmethod
    def iter_required_fields(cls, model_cls):
        for field in model_cls._meta.get_fields():
            if isinstance(field, (models.AutoField, ManyToManyField)):
                continue
            if not field.null and field.default is models.NOT_PROVIDED:
                yield field.name
    
    
    @classmethod
    def iter_unique_field_combinations(cls, model_cls):
        for unique_constraint_fields in model_cls._meta.unique_together:
            yield unique_constraint_fields
            
        for field in model_cls._meta.get_fields():
            if hasattr(field, 'field'):  # relation
                if isinstance(field, (OneToOneField,
                                      OneToOneRel)):
                    yield (field.name,)  # one-to-one always unique
                    continue
                field = field.field
            if field.unique:
                yield (field.name,)
    
    
    @classmethod
    def get_table_fullname(cls, model_cls):
        return model_cls._meta.db_table
    
    
    @classmethod
    def get_model_cls_by_table_fullname(cls, name, adapter_settings):
        for models in apps.all_models.values():
            for model in models.values():
                if cls.get_table_fullname(model) == name:
                    return model
    
    
    #--- methods for working with items ----------------------------------------
    
    @classmethod
    def get(cls, items_and_fkeys, adapter_settings):
        using = adapter_settings.get('using', 'default')
        
        #--- first getting item models from database ---------------------------
        if not items_and_fkeys:
            return []
        
        all_items_filters = Q()
        getters = items_and_fkeys[0][0].getters
        
        for item, fkeys in items_and_fkeys:
            one_item_filters = Q()
            for group in getters:
                group_filters = Q()
                skip_group_filters = False
                # in case related item was not in database
                for field_name in group:
                    if field_name not in item or \
                            item[field_name] is None:
                        skip_group_filters = True
                        break
                    
                    if field_name in item.fields:
                        field_value = item[field_name]
                        group_filters &= Q(**{field_name: field_value})
                    elif field_name in item.relations:
                        related_models = fkeys.get(field_name)
                        if not related_models:
                            # failed to get or created related model before
                            skip_group_filters = True
                            break
                        
                        relation = item.relations[field_name]
                        if not relation['relation_type'].is_x_to_many():
                            group_filters &= Q(
                                **{field_name: related_models[0]})
                        else:
                            contains_any = Q()
                            for related_model in related_models:
                                contains_any |= Q(**{field_name: related_model})
                            group_filters &= contains_any
                
                if not skip_group_filters:
                    one_item_filters |= group_filters
                
            if one_item_filters.children:
                all_items_filters |= one_item_filters
        
        return item.model_cls.objects.db_manager(using) \
            .filter(all_items_filters).all()
    
    @classmethod
    def create_blank_model(cls, model_cls, adapter_settings):
        return model_cls()
    
    @classmethod
    def add_related_models(cls, model, fkey, related_models):
        related_list = getattr(model, fkey)
        all_models = related_list.all()
        for related_model in related_models:
            if related_model not in all_models:
                related_list.add(related_model)
                
    @classmethod
    def clear_related_models(cls, model, fkey):
        getattr(model, fkey).clear()  
    
    @classmethod
    def save_model(cls, model, adapter_settings):
        using = adapter_settings.get('using', 'default')
        model.save(using=using)
    
    #--- helper functions ------------------------------------------------------
    
    @classmethod
    def get_primary_key_names(cls, model_cls):
        return (model_cls._meta.pk.name,)
    
    @classmethod
    def get_related_x_to_many(cls, model, field_name):
        return list(getattr(model, field_name).all())
    
    
    #--- methods for tests -----------------------------------------------------
    
    @staticmethod
    def get_all_models(model_cls, adapter_settings, sort_key=None):
        models = list(model_cls.objects.db_manager(
            adapter_settings.get('using', 'default')).all())
        if sort_key:
            models.sort(key=sort_key)
        return models


from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase



class TestBulkSlicing(TestBase):
    """ Tests for bulk item slicing. """
    
    
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    def test_bulk_item_slice(self):
        
        class ItemsGeneralOne(Item):
            model_cls = self.ModelGeneralOne
                
        class ItemsGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
                
        
        bulk = ItemsGeneralOne.Bulk()
        bulk['f_string'] = 'str-value'
        
        items = []
        for i in range(10):
            item = ItemsGeneralOne(f_integer=i)
            items.append(item)
            bulk.add(item)
        
        first_slice = bulk.slice(0, 5)
        self.assertEqual(items[:5], first_slice.bulk)
        self.assertEqual(first_slice['f_string'], 'str-value')
        
        first_slice['f_text'] = 'text-value'
        self.assertNotIn('f_text', bulk)
        
        second_slice = bulk.slice(start=0, step=2)
        self.assertEqual(items[0::2], second_slice.bulk)
        self.assertEqual(second_slice['f_string'], 'str-value')
        
        bulk['f_float'] = '10.10'
        self.assertNotIn('f_float', second_slice)


""" This module contains exceptions for
:py:class:`~save_to_db.core.item_base.ItemBase` related to fields usage.
"""


class ItemFieldError(ValueError):
    """ General exception for :py:class:`~save_to_db.core.item_base.ItemBase`
    field usage.
    """




import os
import django
from django.core.management import call_command

from save_to_db.adapters.django_adapter import DjangoAdapter
from save_to_db.core.persister import Persister

from . import project
from .project import settings
from .django_test_base import DjangoTestBase

from .. import register_test
from ..test_fields_and_relations import TestFieldsAndRelations
from ..test_constraints import TestConstraints
from ..test_complete_item_metaclass import TestCompleteItemMetaclass
from ..test_item_cls_manager import TestItemClsManager
from ..test_item_fields import TestItemFields
from ..test_item_hooks import TestItemHooks
from ..test_item_persist import TestItemPersist
from ..test_item_dump import TestItemDump
from ..test_bulk_slicing import TestBulkSlicing
from ..test_scope import TestScope


# setting up Django and loading models
settings.INSTALLED_APPS = tuple(
    '{}.{}'.format(project.__name__, app) for app in settings.INSTALLED_APPS)
os.environ.setdefault(
    'DJANGO_SETTINGS_MODULE',
    '{}.settings'.format(project.__name__))
django.setup()
call_command('migrate', verbosity=0)
from .project.orm_only import models



AdapterPrefix = 'Django'
persister = Persister(DjangoAdapter,
                      adapter_settings={})

models = {
    'ModelFieldTypes': models.ModelFieldTypes,
    
    'ModelGeneralOne': models.ModelGeneralOne,
    'ModelGeneralTwo': models.ModelGeneralTwo,

    'ModelConstraintsOne': models.ModelConstraintsOne,
    'ModelConstraintsTwo': models.ModelConstraintsTwo,
    'ModelConstraintsThree': models.ModelConstraintsThree,
    'ModelConstraintsFour': models.ModelConstraintsFour,
    'ModelConstraintsFive': models.ModelConstraintsFive,
    'ModelConstraintsSix': models.ModelConstraintsSix,
    'ModelConstraintsSelf': models.ModelConstraintsSelf,
    
    'ModelInvalidFieldNames': models.ModelInvalidFieldNames,

    'ModelManyRefsOne': models.ModelManyRefsOne,
    'ModelManyRefsTwo': models.ModelManyRefsTwo,
}

for test_cls in (TestFieldsAndRelations,
                 TestConstraints,
                 TestItemClsManager,
                 TestCompleteItemMetaclass,
                 TestItemFields,
                 TestItemHooks,
                 TestItemPersist,
                 TestItemDump,
                 TestBulkSlicing,
                 TestScope):
    
    register_test(AdapterPrefix, DjangoTestBase,
                  test_cls, models, persister)